%
% @(#) built.pl
% Some built-in predicates for Prolog Cafe
% 

% @author Mutsunori Banbara 
%         banbara@pascal.seg.kobe-u.ac.jp
%         Nara National College of Technology
% @author Naoyuki Tamura 
%         tamura@kobe-u.ac.jp
%         Faculty of Engineering, Kobe University

:- dynamic '$current_predicates'/2.

/*--------------------------------------------------------------
	Version
--------------------------------------------------------------*/
version :- '$version'(X), write(X),nl, !.

/*--------------------------------------------------------------
	Read in Programs
--------------------------------------------------------------*/
fcompile(Files) :- var(Files), !, fail.
fcompile([]) :- !.
fcompile(File) :- atom(File), !,
	'$fcompile'(File).
fcompile([File|Files]) :- atom(File), !,
	'$fcompile'(File),
	fcompile(Files).

'$fcompile'(File) :- plcs(File), !.

compile(Files) :- fcompile(Files).

/*--------------------------------------------------------------
	Calling C# from Prolog (P# Specific)
--------------------------------------------------------------*/

load_assembly(X) :- load_assembly(X).

cs_constructor(X,Y) :- cs_constructor(X,Y).

cs_method(X,Y,Z) :- cs_method(X,Y,Z).

cs_get_field(X,Y,Z) :- cs_get_field(X,Y,Z).

cs_set_field(X,Y,Z) :- cs_set_field(X,Y,Z).

cs_term(X,Y) :- cs_term(X,Y).

url_source(X,Y) :- url_source(X,Y).

/*--------------------------------------------------------------
	Using Tables (Prolog Cafe Specific)
--------------------------------------------------------------*/
open_table(X) :- open_table(X).

close_table(X) :- close_table(X).
%close_table(X) :- '$close_table'(X), !.
%close_table(X) :- '$illarg'(type(table), close_table(X), 1).

clear_table(X) :- clear_table(X).
%clear_table(X) :- '$clear_table'(X), !.
%clear_table(X) :- '$illarg'(type(table), clear_table(X), 1).

set_table(X) :- set_table(X).
%set_table(X) :- '$set_table'(X), !.
%set_table(X) :- '$illarg'(type(table), set_table(X), 1).

current_table(X) :- current_table(X).

get_term(Key, Term) :- ground(Key), !, 
	'$get_term'(Key, Term).
get_term(Key, Term) :- nonvar(Key), !, 
	'$illarg'(type(ground), get_term(Key, Term), 1).
get_term(Key, Term) :- 
	'$illarg'(var, get_term(Key, Term), 1).

get_term(Table, Key, Term) :- ground(Key), !,
	'$get_term'(Table, Key, Term).
get_term(Table, Key, Term) :- nonvar(Key), !,
	'$illarg'(type(ground), get_term(Table, Key, Term), 2).
get_term(Table, Key, Term) :- 
	'$illarg'(var, get_term(Table, Key, Term), 2).

put_term(Key, Term) :- ground(Key), !,
	'$put_term'(Key, Term).
put_term(Key, Term) :- nonvar(Key), !,
	'$illarg'(type(ground), put_term(Key, Term), 1).
put_term(Key, Term) :- 
	'$illarg'(var, put_term(Key, Term), 1).

put_term(Table, Key, Term) :- ground(Key), !,
	'$put_term'(Table, Key, Term).
put_term(Table, Key, Term) :- nonvar(Key), !,
	'$illarg'(type(ground), put_term(Table, Key, Term), 2).
put_term(Table, Key, Term) :- 
	'$illarg'(var, put_term(Table, Key, Term), 2).

/*--------------------------------------------------------------
	Arithmetic
--------------------------------------------------------------*/
Z is Y :- Z is Y.

X =:= Y :- X =:= Y.

X =\= Y :- X =\= Y.

X < Y :- X < Y.

X > Y :- X > Y.

X =< Y :- X =< Y.

X >= Y :- X >= Y.

/*--------------------------------------------------------------
	Comparison of Terms
--------------------------------------------------------------*/
X == Y :- X == Y.

X \== Y :- X \== Y.

X @< Y :- X @< Y.

X @> Y :- X @> Y.

X @=< Y :- X @=< Y.

X @>= Y :- X @>= Y.

?=(X, Y) :- ?=(X, Y).

compare(X,Y,Z) :- compare(X,Y,Z).

sort([],[]) :- !.
sort([H|T], Result) :-
        '$partition'(H, T, S, G) ,
        sort(S, LS),
        sort(G, LG) ,
        '$append'(LS, [H|LG], Result).

'$partition'(K, [], [], []) :- !.
'$partition'(K, [H|T], [H|S], G) :- H @< K, !, 
	'$partition'(K,T,S,G).
'$partition'(K, [H|T], S, G) :- H == K, !,
	'$partition'(K,T,S,G).
'$partition'(K, [H|T], S, [H|G]) :- 
	'$partition'(K, T, S, G).

/*--------------------------------------------------------------
	Control
--------------------------------------------------------------*/
(P, Q) :- call(P), call(Q).

!.

\+(P) :- call(P), !, fail.
\+(P).

(IF -> THEN) :- '$if_then_else'(IF, THEN, fail).

(IF -> THEN; ELSE) :- !, '$if_then_else'(IF, THEN, ELSE).

'$if_then_else'(IF, THEN, ELSE) :- call(IF), !, call(THEN).
'$if_then_else'(IF, THEN, ELSE) :- call(ELSE).

(P; Q) :- call(P).
(P; Q) :- call(Q).

otherwise.
true.

false :- false.
fail  :- fail.

repeat.
repeat :- repeat.

call(G) :- '$call'(G).

%call(G) :- 
%	functor(G,F,A), 
%	clause('$dynamic_predicates'(F,A),_), 
%	!,
%	'$call'(G).
%call(G) :- '$call'(G).
%call(G) :- clause(G, G1), call(G1).

once(G) :- call(G), !.

/*--------------------------------------------------------------
	Meta-Logic
--------------------------------------------------------------*/
var(X) :- var(X).

nonvar(X) :- nonvar(X).

atom(X) :- atom(X).

float(X) :- fail.

integer(X) :- integer(X).

number(X) :- integer(X), !.
number(X) :- float(X).

atomic(X) :- atomic(X).

cs_object(X) :- cs_object(X).

ground(X) :- var(X), !, fail.
ground(X) :- atomic(X), !.
ground(X) :- functor(X, F, N), '$ground'(1, N, X).

'$ground'(I, N, X) :- I > N, !.
'$ground'(I, N, X) :- 
	arg(I, X, A),
	ground(A),
	I1 is I + 1,
	'$ground'(I1, N, X).

callable(X) :- atom(X), !.
callable(X) :- compound(X).

compound(X) :- nonvar(X), functor(X, _, A), A > 0.

functor(X, Y, Z) :- nonvar(X), !, functor(X, Y, Z).
functor(X, Y, Z) :- integer(Z), !,
	(    atomic(Y) -> functor(X, Y, Z)
	;    '$illarg'(type(atomic), functor(X,Y,Z), 2)
	).
functor(X, Y, Z) :- nonvar(Z), !,
	'$illarg'(type(integer), functor(X,Y,Z), 3).
functor(X, Y, Z) :- 
	'$illarg'(var, functor(X,Y,Z), 3).

arg(X, Y, Z) :- integer(X), !,
	(    nonvar(Y) -> arg(X, Y, Z)
	;    '$illarg'(var, arg(X,Y,Z), 2)
	).
arg(X, Y, Z) :- nonvar(X), !,
	'$illarg'(type(integer), arg(X,Y,Z), 1).
arg(X, Y, Z) :- 
	'$illarg'(var, arg(X,Y,Z), 1).

X =.. Y :-  nonvar(X), !,
	functor(X, Name, Arity),
	'$term_to_list'(X, 1, Arity, As),
	Y = [Name|As].
X =.. Y :- nonvar(Y), !, '$univ_internal'(X,Y).
X =.. Y :- '$illarg'(var, (X =.. Y), 2).

'$univ_internal'(X, Y) :- '$list_to_term'(Y, X), !.
'$univ_internal'(X, Y) :-  '$illarg'(var, (X =.. Y), 2).

'$term_to_list'(X, I, N, []) :- I > N, !.
'$term_to_list'(X, I, N, [A|L]) :-
	arg(I, X, A),
	I1 is I + 1,
	'$term_to_list'(X, I1, N, L).

'$list_to_term'([F|Args], Term) :-
	length(Args, N),
	functor(Term, F, N),
	'$list_to_term'(Args, Term, 1).

'$list_to_term'(Args, _, _) :- var(Args), !, fail.
'$list_to_term'([], _, _).
'$list_to_term'([A|Args], Term, I) :-
	arg(I, Term, A),
	I1 is I + 1,
	'$list_to_term'(Args, Term, I1).


atom_chars(Atom, Chars) :- atom(Atom), !, 
	'$atom_chars'(Atom, Chars).
atom_chars(Atom, Chars) :- nonvar(Atom), !, 
	'$illarg'(type(atom), atom_chars(Atom,Chars), 1).
atom_chars(Atom, Chars) :- var(Atom),
	'$atom_chars'(Atom0, Chars), !, Atom = Atom0.
atom_chars(Atom, Chars) :-
	'$illarg'(type(list(char)), atom_chars(Atom,Chars), 2).

number_chars(Number, Chars) :- number(Number), !,
	'$number_chars'(Number, Chars).
number_chars(Number, Chars) :- nonvar(Number), !,
	'$illarg'(type(number), number_chars(Number,Chars), 1).
number_chars(Number, Chars) :- var(Number),
	'$number_chars'(Number0, Chars), !, Number = Number0.
number_chars(Number, Chars) :- 
	'$atom_chars'(_, Chars), !, fail.
number_chars(Number, Chars) :- 
	'$illarg'(type(list(char)), number_chars(Number,Chars), 2).

name(Atomic, Chars) :- number(Atomic), !,
	'$number_chars'(Atomic, Chars).
name(Atomic, Chars) :- atomic(Atomic), !,
	'$atom_chars'(Atomic, Chars).
name(Atomic, Chars) :- nonvar(Atomic), !,
	'$illarg'(type(atomic), name(Atomic,Chars), 1).
name(Atomic, Chars) :- var(Atomic),
	'$number_chars'(Atomic0, Chars), !, Atomic = Atomic0.
name(Atomic, Chars) :- var(Atomic),
	'$atom_chars'(Atomic0, Chars), !, Atomic = Atomic0.
name(Atomic, Chars) :-
	'$illarg'(type(list(char)), name(Atomic,Chars), 2).

/*--------------------------------------------------------------
	Modification of the Program
--------------------------------------------------------------*/
assert(Clause) :- assertz(user, Clause). 
assert(Table, Clause) :- assertz(Table, Clause). 

asserta(Clause) :- asserta(user, Clause).
assertz(Clause) :- assertz(user, Clause).
clause(Head, Body) :- clause(user, Head, Body).
retract(Clause) :- retract(user, Clause).
retractall(Head) :- retractall(user, Head).
abolish(X) :- abolish(user, X).

%asserta(Table, Clause) :-
%	'$canonical_clause'(Clause, Key, Clause0),
%	get_term(Table, Key, Cls0),
%	copy_term([Clause0|Cls0], Cls),
%	put_term(Table, Key, Cls).

%assertz(Table, Clause) :-
%	'$canonical_clause'(Clause, Key, Clause0),
%	get_term(Table, Key, Cls0),
%	'$append'(Cls0, [Clause0], Cls1),
%	copy_term(Cls1, Cls),
%       put_term(Table, Key, Cls).

assertz(Table, Clause) :-
	'$canonical_clause'(Clause, Key, Clause0),
	'$assert_current_predicate'(Table, Key),
	'$get_term'(Table, Key, Cls0),
	copy_term([Clause0|Cls0], Cls),
	'$put_term'(Table, Key, Cls).

asserta(Table, Clause) :-
	'$canonical_clause'(Clause, Key, Clause0),
	'$assert_current_predicate'(Table, Key),
	'$get_term'(Table, Key, Cls0),
	'$append'(Cls0, [Clause0], Cls1),
	copy_term(Cls1, Cls),
	'$put_term'(Table, Key, Cls).

%clause(Table, Head, Body) :-
%	'$canonical_clause'((Head :- Body), Key, C),
%	get_term(Table, Key, Cls0),
%	copy_term(Cls0, Cls),
%	'$member'(Cl, Cls),
%	C = Cl.

clause(Table, Head, Body) :-
	'$canonical_clause'((Head :- Body), Key, C),
	'$get_term'(Table, Key, Cls0),
	copy_term(Cls0, Cls),
	'$member_in_reverse'(Cl, Cls),
	C = Cl.

%retract(Table, Clause) :-
%	'$canonical_clause'(Clause, Key, Cl),
%	get_term(Table, Key, Cls0),
%	copy_term(Cls0, Cls),
%	'$select'(C, Cls, Cls1),
%	C = Cl,
%	!,
%	put_term(Table, Key, Cls1).

retract(Table, Clause) :-
	'$canonical_clause'(Clause, Key, Cl),
	'$get_term'(Table, Key, Cls0),
	copy_term(Cls0, Cls),
	'$select_in_reverse'(C, Cls, Cls1),
	C = Cl,
	'$put_term'(Table, Key, Cls1).

retractall(Table, Head) :- nonvar(Head), !,
	'$functor'(Head, F, A),
	'$get_term'(Table, F/A, Cls0),
	copy_term(Cls0, Cls1),
	'$retractall'(Head, Cls1, Cls),
	'$put_term'(Table, F/A, Cls).
retractall(Table, Head) :- 
	'$illarg'(var, retractall(Table, Head), 2).
	
'$retractall'(_, [], []) :- !.
'$retractall'(H, [Cl|Cls], Cls1) :-  
	copy_term(H, H1),
	Cl = (H1 :- _), 
	!,
	'$retractall'(H, Cls, Cls1).
'$retractall'(H, [Cl|Cls], [Cl|Cls1]) :- 
	'$retractall'(H, Cls, Cls1).

abolish(Table, F/A) :- atom(F), integer(A), !,
	'$put_term'(Table, F/A, []).
abolish(Table, X) :- 
	'$illarg'(type(functor/arity), abolish(Table, X), 2).

'$canonical_clause'((H :- B), F/A, (H :- B)) :- !,
	functor(H,F,A),
	atom(F).
'$canonical_clause'(H, F/A, (H :- true)) :-
	functor(H,F,A),
	atom(F).

'$assert_current_predicate'(Table, F/A) :-
	clause(Table, '$current_predicates'(F,A), _),
	!.
'$assert_current_predicate'(Table, F/A) :-
	'$canonical_clause'('$current_predicates'(F,A), Key, Cl),
	'$get_term'(Table, Key, Cls0),
	copy_term([Cl|Cls0], Cls),
	'$put_term'(Table, Key, Cls).

/*--------------------------------------------------------------
	All Solutions
--------------------------------------------------------------*/
findall(Template, Goal, Bag) :-
	open_table(Table),
	'$findall'(Table, Template, Goal, Bag),
	close_table(Table).

'$findall'(Table, X, Goal, Bag) :-
	call(Goal), 
	'$get_term'(Table, all, Y),
	copy_term([X|Y], Z),
	'$put_term'(Table, all, Z),
	fail.
'$findall'(Table, X, _, Bag) :-
	'$get_term'(Table, all, Y),
	copy_term(Y, Z),
	'$reverse'(Z, Bag),
	!.

/*--------------------------------------------------------------
	DCG (Definite Clause Grammers)
--------------------------------------------------------------*/
'C'([X|S], X, S).

/* term expansion */
expand_term(T1, T3) :-	
        '$expand_term_internal'(T1, T2),
	T2 = T3.

'$expand_term_internal'(T1, T2) :- var(T1), !, T2 = T1.
'$expand_term_internal'(T1, T2) :- '$dcg_expansion'(T1, T2), !.
'$expand_term_internal'(T, T).

%'$expand_term_internal'(T1, T2) :-
%	(   var(T1) -> T2 = T1
%	;   '$dcg_expansion'(T1, T2) -> true
%	;   T2 = T1
%	).

'$dcg_expansion'(M:Rule, M:Clause) :- !,
	nonvar(Rule),
	'$dcg_expansion'(Rule, Clause).
'$dcg_expansion'((Head-->B), (H1:-B1,B2)) :-
	nonvar(Head),
	Head = (H,List), !,
	'$dcg_translate_dcg_atom'(H, H1, S0, S1),
	'$dcg_translate_dcg'(B, B1, S0, S),
	'$dcg_translate_dcg'(List, B2, S1, S).
'$dcg_expansion'((H-->B), (H1:-B1)) :-
	'$dcg_translate_dcg_atom'(H, H1, S0, S),
	'$dcg_translate_dcg'(B, B1, S0, S).

'$dcg_translate_dcg'(X, Y, S0, S) :-
	'$dcg_translate_dcg'(X, Y0, Tail, S0, S),
	(   Tail\==S0 -> S=Tail, '$dcg_connect'(X, Y0, Y)
	;   '$dcg_connect'(X, Y0, S=Tail, Y)
	).

'$dcg_connect'((_->_), X, X) :- X = (_->_), !.
'$dcg_connect'(_, (P->Q), ((P->Q),true)) :- !.
'$dcg_connect'(_, X, X).

'$dcg_connect'((_->_), (P->Q0), Eq, (P->Q)) :- !, '$dcg_and'(Q0, Eq, Q).
'$dcg_connect'(_, (P->Q), Eq, ((P->Q),Eq)) :- !.
'$dcg_connect'(_, X0, Eq, X) :- '$dcg_and'(X0, Eq, X).

'$dcg_translate_dcg'(X, X1, S, S0, S) :- var(X), !,
	'$dcg_translate_dcg_atom'(X, X1, S0, S).
'$dcg_translate_dcg'(M:X, M:Y, Tail, S0, S) :- !,
	'$dcg_translate_dcg'(X, Y, Tail, S0, S).
'$dcg_translate_dcg'([], true, S0, S0, _) :- !.
'$dcg_translate_dcg'([X|Y], XandY, Tail, S0, S) :- !,
	'$dcg_translate_dcg'(Y, Y1, Tail, S1, S),
	'$dcg_and'('C'(S0,X,S1), Y1, XandY).
'$dcg_translate_dcg'(\+X, (X1 -> fail; S=S0), S, S0, S) :- !,
	'$dcg_translate_dcg'(X, X1, S1, S0, S1).
'$dcg_translate_dcg'((X,Y), XandY, Tail, S0, S) :- !,
	'$dcg_translate_dcg'(X, X1, S1, S0, S1),
	'$dcg_translate_dcg'(Y, Y1, Tail, S1, S),
	'$dcg_and'(X1, Y1, XandY).
'$dcg_translate_dcg'((X->Y), (X1->Y1), Tail, S0, S) :- !,
	'$dcg_translate_dcg'(X, X1, S1, S0, S1),
	'$dcg_translate_dcg'(Y, Y1, Tail, S1, S).
'$dcg_translate_dcg'(if(X,Y,Z), if(X1,Y1,Z1), S, S0, S) :- !,
	'$dcg_translate_dcg'(X, X1, S0, S1),
	'$dcg_translate_dcg'(Y, Y1, S1, S),
	'$dcg_translate_dcg'(Z, Z1, S1, S).
'$dcg_translate_dcg'((X;Y), (X1;Y1), S, S0, S) :- !,
	'$dcg_translate_dcg'(X, X1, S0, S),
	'$dcg_translate_dcg'(Y, Y1, S0, S).
'$dcg_translate_dcg'(!, !, S0, S0, _) :- !.
'$dcg_translate_dcg'({G}, call(G), S0, S0, _) :- var(G), !.
'$dcg_translate_dcg'({G}, G, S0, S0, _) :- !.
'$dcg_translate_dcg'(X, X1, S, S0, S) :-
	'$dcg_translate_dcg_atom'(X, X1, S0, S).

'$dcg_and'(X, Y, Z) :- X==true, !, Z=Y.
'$dcg_and'(X, Y, Z) :- Y==true, !, Z=X.
'$dcg_and'(X, Y, (X,Y)).

'$dcg_translate_dcg_atom'(X, X1, S0, S) :- var(X), !,
	X1 = phrase(X,S0,S).
'$dcg_translate_dcg_atom'(M:X, M:X1, S0, S) :- !,
	'$dcg_translate_dcg_atom'(X, X1, S0, S).
'$dcg_translate_dcg_atom'(X, X1, S0, S) :-
	functor(X, F, A),
	A1 is A+1,
	A2 is A+2,
	functor(X1, F, A2),
	arg(A1, X1, S0),
	arg(A2, X1, S),
	'$dcg_copy_args'(A, X, X1).

'$dcg_copy_args'(0, _, _) :- !.
'$dcg_copy_args'(I, F1, F2) :-
	arg(I, F1, X),
	arg(I, F2, X),
	I1 is I-1,
	'$dcg_copy_args'(I1, F1, F2).

/*--------------------------------------------------------------
	Miscellaneous
--------------------------------------------------------------*/
time(Goal) :- 
	statistics(runtime, _),
	once(Goal), 
	statistics(runtime, [_, T]),
	write('RUN time = '),  write(T), write(' msec. '), nl.

X = Y :- X = Y.

length(L, N) :- var(N), !, '$length'(L, 0, N).
length(L, N) :- '$dlength'(L, 0, N).

'$length'([], I, I).
'$length'([_|L], I0, I) :- I1 is I0+1, '$length'(L, I1, I).

'$dlength'([], I, I) :- !.
'$dlength'([_|L], I0, I) :- I0<I, I1 is I0+1, '$dlength'(L, I1, I).

halt :- halt.

abort :- abort.

copy_term(X, Y) :- copy_term(X, Y).

%copy_term(X, Z) :-
%	'$copy_term'(X, Z, [], _).
%
%'$copy_term'(X, Z, Vs, Vs) :-
%	var(X), '$member'(X1=Z, Vs), X == X1, !.
%'$copy_term'(X, Z, Vs0, [X=Z|Vs0]) :- var(X), !.
%'$copy_term'(X, X, Vs, Vs) :- ground(X), !.
%'$copy_term'(X, Z, Vs0, Vs) :-
%	functor(X, F, N),
%	functor(Z, F, N),
%	'$copy_term_args'(1, N, X, Z, Vs0, Vs).
%
%'$copy_term_args'(I, N, _, _, Vs, Vs) :- I > N, !.
%'$copy_term_args'(I, N, X, Z, Vs0, Vs) :-
%	arg(I, X, X1),
%	arg(I, Z, Z1),
%	'$copy_term'(X1, Z1, Vs0, Vs1),
%	I1 is I + 1,
%	'$copy_term_args'(I1, N, X, Z, Vs1, Vs).

numbervars(X, VI, VN) :- integer(VI), VI >= 0, !,
	'$numbervars'(X, VI, VN).

'$numbervars'(X, VI, VN) :- var(X), !,
	X = '$VAR'(VI),		% This structure is checked in write
	VN is VI + 1.
'$numbervars'(X, VI, VI) :- atomic(X), !.
'$numbervars'(X, VI, VI) :- cs_object(X), !. % changed by JJC
'$numbervars'(X, VI, VN) :-
	functor(X, F, N),
	'$numbervars_str'(1, N, X, VI, VN).

'$numbervars_str'(I, I, X, VI, VN) :- !,
	arg(I, X, A),
	'$numbervars'(A, VI, VN).
'$numbervars_str'(I, N, X, VI, VN) :-
	arg(I, X, A),
	'$numbervars'(A, VI, VN1),
	I1 is I + 1,
	'$numbervars_str'(I1, N, X, VN1, VN).

statistics :-
	findall(M, '$statistics_mode'(M), Areas),
	'$member'(Area, Areas),
	'$statistics'(Area, [Used, Free]),
	'$write_statistics'(Area, [Used, Free]),
	fail.
statistics.

'$write_statistics'(runtime, [Start, Previous]) :- !,
	write('runtime = '), 
	write(Start), 
	write(' from start, '),
	write(Previous), 
	write(' from previous'),
	nl.
'$write_statistics'(Area, [Used, Free]) :-
	Total is Used + Free,
	write(Area), 
	write(' = '), 
	write(Total), 
	write(' entries, '),
	write(Used), 
	write(' in use, '),
	write(Free), 
	write(' in free'), 
	nl.

statistics(Mode, Result) :- 
	nonvar(Mode),
	'$statistics_mode'(Mode),
	!,
	'$statistics'(Mode, Result).
statistics(Mode, Result) :- 
	findall(M, '$statistics_mode'(M), Domain),
	'$illarg'(domain(atom,Domain), statistics(Mode, Result), 1).

'$statistics_mode'(runtime).
'$statistics_mode'(trail).
'$statistics_mode'(choice).

/*--------------------------------------------------------------
	Error
--------------------------------------------------------------*/
'$illarg'(var, Goal, ArgNo) :- 
	'$print_error'(instantiation_error(Goal, ArgNo)).

'$illarg'(type(TypeName), Goal, ArgNo) :- 
	arg(ArgNo, Goal, Arg),
	(  nonvar(Arg) -> 
	   Error = type_error(Goal, ArgNo, TypeName, Arg)
	;  Error = instantiation_error(Goal, ArgNo)
	),
	'$print_error'(Error).

'$illarg'(domain(TypeName,ExpDomain), Goal, ArgNo) :-
	arg(ArgNo, Goal, Arg),
	(  '$right_type'(TypeName, Arg) ->
	   Error = domain_error(Goal, ArgNo, ExpDomain, Arg)
	;  nonvar(Arg) ->
	   Error = type_error(Goal, ArgNo, TypeName, Arg)
	;  Error = instantiation_error(Goal, ArgNo)
	),
	'$print_error'(Error).

'$illarg'(context(ContextType,CommandType), Goal, _) :-
	'$print_error'(context_error(Goal,ContextType,CommandType)).

'$illarg'(existence(ObjType,Culprit,Message), Goal, ArgNo) :-
	'$print_error'(existence_error(Goal,ArgNo,ObjType,Culprit,Message)).

'$illarg'(permission(Operation, ObjType, Culprit, Message), Goal, _) :-
	'$print_error'(permission_error(Goal,Operation,ObjType,Culprit,Message)).

'$illarg'(_, _, _) :- 
	write('Error Occurs !!'), nl,
	abort.

'$right_type'(term, _).
'$right_type'(atom, A) :- atom(A).
'$right_type'(atomic, A) :- atomic(A).
'$right_type'(number, N) :- number(N).
'$right_type'(integer, I) :- integer(I).
'$right_type'(float, I) :- float(I).

'$print_error'(instantiation_error(Goal, 0)) :- !,
	write('{INSTANTIATION ERROR: in '),
	write(Goal),
	write('}'), nl,
	abort.

'$print_error'(instantiation_error(Goal, ArgNo)) :- !,
	write('{INSTANTIATION ERROR: in '),
	write(Goal),
	write(' - arg '),
	write(ArgNo),
	write('}'), nl,
	abort.

'$print_error'(type_error(evaluation, 0, expression, Culprit)) :- !,
	write('{TYPE ERROR: expected expression, found '),
	write(Culprit),
	write('}'), nl,
	abort.

'$print_error'(type_error(Goal, ArgNo, TypeName, _)) :- var(TypeName), !,
	write('{TYPE ERROR: '),
	write(Goal),
	write(' - arg '),
	write(ArgNo),
	write('}'), nl,
	abort.

'$print_error'(type_error(Goal, ArgNo, TypeName, Culprit)) :- !,
	write('{TYPE ERROR: '),
	write(Goal),
	write(' - arg '),
	write(ArgNo),
	write(': expected '),
	write(TypeName),
	write(', found '),
	write(Culprit),
	write('}'), nl,
	abort.

'$print_error'(domain_error(Goal,0,_,_)) :- float(Goal), !, 
	write('{DOMAIN ERROR: expected finite float, found '),
	write(Goal),
	write('}'), nl,
	abort.

'$print_error'(domain_error(Goal,ArgNo,TypeName,_)) :- var(TypeName), !, 
	write('{DOMAIN ERROR: '),
	write(Goal),
	write(' - arg '),
	write(ArgNo),
	write('}'), nl,
	abort.

'$print_error'(domain_error(Goal,ArgNo,TypeName,Culprit)) :- !, 
	write('{DOMAIN ERROR: '),
	write(Goal),
	write(' - arg '),
	write(ArgNo),
	write(': expected '),
	write(TypeName),
	write(', found '),
	write(Culprit),
	write('}'), nl,
	abort.

'$print_error'(existence_error(Goal, 0, ObjType, Culprit, _Message)) :- !,
	write('{EXISTENCE ERROR: '),
	write(Goal), 
	write(' '),
	write(ObjType), 
	write(' '),
	write(Culprit), 
	write(' does not exist'),
	write('}'), nl,
	abort.

'$print_error'(existence_error(Goal, ArgNo, ObjType, Culprit, _Message)) :- !,
	write('{EXISTENCE ERROR: '),
	write(Goal), 
	write(' - arg '),
	write(ArgNo),
	write(': '),
	write(ObjType), 
	write(' '),
	write(Culprit), 
	write(' does not exist'),
	write('}'), nl,
	abort.

'$print_error'(context_error(Goal, ContextType, CommandType)) :- !, 
	write('{CONTEXT ERROR: '),
	write(Goal), 
	write(' -  '),
	write(CommandType),
	write(' appeared in '),
	write(ContextType),
	write('}'), nl,
	abort.

'$print_error'(permission_error(Goal, Operation, ObjType, Culprit, _Message)) :- !, 
	write('{PERMISSION ERROR: '),
	write(Goal), 
	write(' - can not '),
	write(Operation),
	write(' '),
	write(ObjType), 
	write(' '),
	write(Culprit), 
	write('}'), nl,
	abort.

'$print_error'(system_error(Msg)) :- !, 
	write('{SYSTEM ERROR: '),
	write(Msg), 
	write('}'), nl,
	abort.

'$print_error'(Error) :-
	write('{ERROR: '),
	write(Error), 
	write('}'), nl,
	abort.

%illarg(var, Goal, ArgNo) :-
%	raise_exception(instantiation_error(Goal,ArgNo)).
%illarg(domain(ExpType,ExpDomain), Goal, ArgNo) :-
%	arg(ArgNo, Goal, Arg),
%	(   type_check(ExpType, Arg) ->
%	    Error = domain_error(Goal,ArgNo,ExpDomain,Arg)
%	;   nonvar(Arg) ->
%	    Error = type_error(Goal,ArgNo,ExpType,Arg)
%	;   Error = instantiation_error(Goal,ArgNo)
%	),
%	raise_exception(Error).
%illarg(type(ExpType), Goal, ArgNo) :-
%	arg(ArgNo, Goal, Arg),
%	(   nonvar(Arg) ->
%	    Error = type_error(Goal,ArgNo,ExpType,Arg)
%	;   Error = instantiation_error(Goal,ArgNo)
%	),
%	raise_exception(Error).
%illarg(context(ContextType,CommandType), Goal, _) :-
%	raise_exception(context_error(Goal,ContextType,CommandType)).
%illarg(existence(ObjType,Culprit,Message), Goal, ArgNo) :-
%	raise_exception(existence_error(Goal,ArgNo,ObjType,Culprit,Message)).
%illarg(permission(Operation,ObjType,Culprit,Message), Goal, _) :-
%	raise_exception(permission_error(Goal,Operation,ObjType,Culprit,Message)).
%
%type_check(term, _).
%type_check(atom, A) :- atom(A).
%type_check(atomic, A) :- atomic(A).
%type_check(number, N) :- number(N).
%type_check(integer, I) :- integer(I).
%type_check(float, I) :- float(I).

/*--------------------------------------------------------------
	Utilities
--------------------------------------------------------------*/
'$member'(X, [X|_]).
'$member'(X, [_|L]) :- '$member'(X, L).

'$member_in_reverse'(X, [_|L]) :- '$member_in_reverse'(X, L).
'$member_in_reverse'(X, [X|_]).

'$append'([], Zs, Zs).
'$append'([X|Xs], Ys, [X|Zs]) :- '$append'(Xs, Ys, Zs).

'$select'(X,[X|Xs], Xs).
'$select'(X,[Y|Ys], [Y|Zs]) :- '$select'(X,Ys,Zs).

'$select_in_reverse'(X,[Y|Ys], [Y|Zs]) :- '$select_in_reverse'(X,Ys,Zs).
'$select_in_reverse'(X,[X|Xs], Xs).

'$reverse'(Xs, Zs) :- '$reverse'(Xs, [], Zs).

'$reverse'([], Zs, Zs).
'$reverse'([X|Xs], Ys, Zs) :- '$reverse'(Xs, [X|Ys], Zs).

'$list_to_string'([], S) :- !, S = ''.
'$list_to_string'([X|Xs], S) :- 
	name(X, L1),
	'$list_to_string'(Xs, S1),
	name(S1, L2),
	'$append'(L1, L2, L3),
	name(S, L3).

%'$message'([]) :- !, nl.
%'$message'([M|Ms]) :- write(M), write(' '), '$message'(Ms).








