/*
 *  @(#)dollar_arg_3.cs
 *  $B=R8l(B $arg/3
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	public sealed class dollar_arg_3 : Predicate {
		Term arg1, arg2, arg3;

		public dollar_arg_3(Term a1, Term a2, Term a3, Predicate cont) {
			arg1 = a1;
			arg2 = a2;
			arg3 = a3;
			this.cont = cont;
		}
		public dollar_arg_3(){}
		public override void setArgument(Term[] args, Predicate cont){
			arg1 = args[0];
			arg2 = args[1];
			arg3 = args[2];
			this.cont = cont;
		}

		public override Predicate exec() {
			Term a1, a2, a3;
			a1 = arg1.dereference();
			a2 = arg2.dereference();
			a3 = arg3.dereference();

			Term[] args;
			int num, arity;

			if( !a1.isInteger() ) {
				return engine.abort("{TYPE ERROR: " + this + 
					" -arg 1: expected integer, found " + 
					a1 + "}\n");
				//	    return null;
			}
			num = ((IntegerTerm)a1).value();
			if ( a2.isList() ){
				args = new Term[2];
				args[0] = ((ListTerm)a2).car;
				args[1] = ((ListTerm)a2).cdr;
				arity = 2;
			} 
			else if ( a2.isStructure() ){
				args =  ((StructureTerm)a2).args;
				arity = ((StructureTerm)a2).arity;
			} 
			else if ( a2.isVariable() ){
				return engine.abort("{INSTANTIATION ERROR: " + this + " -arg 2}\n");
				//return null;
			}
			else { 
				return engine.abort("{TYPE ERROR: " + this + 
					" -arg 2: expected compound, found " + 
					a2 + "}\n");
				//	    return null;
			}
			if( num < 1 || num > arity ) 
				return engine.fail();
			if( !a3.unify(args[num-1], engine.trail) ) 
				return engine.fail();
			return cont;
		}

		public override int arity() { return 3 ; }

		public override string ToString() {
			return "$arg(" + arg1 + ", " + arg2 + ", " + arg3 + ")";
		}
	}

}