%
% @(#)interpreter.pl
% Small Prolog Interpreter
% 

% @author Mutsunori Banbara 
%         banbara@pascal.seg.kobe-u.ac.jp
%         Nara National College of Technology
% @author Naoyuki Tamura 
%         tamura@kobe-u.ac.jp
%         Faculty of Engineering, Kobe University

%:- dynamic '$consulted_predicates'/2.
%:- dynamic '$consulted_clause'/2.
:- dynamic '$trace_flag'/1.
:- dynamic '$current_leash'/1.
:- dynamic '$current_spy'/2.
:- resource '$ancestor'/2.

%-----------------------------------------------------------
% Small Prolog Interpreter
%-----------------------------------------------------------
cafeteria :-
	'$cafeteria_init',
	write('Small Prolog Interpreter'),nl,
	repeat,
	  ('$trace_flag'(on) -> write('{trace}'),nl; true), %trace mode
	  write('| ?- '),
	  flush_output,
	  read_with_variables(Goal, Vs),
	  '$cafe_execute'(Goal, Vs),
	  Goal == end_of_file,
	  !,
	nl,
	write(bye),
	nl.

%'$cafeteria_init' :- '$consult_init', '$trace_init', !.
'$cafeteria_init' :- '$trace_init', !.

'$cafe_execute'(G, _) :- var(G), !,
	write('Invalid goal'), nl.
'$cafe_execute'(end_of_file, _) :- !.
'$cafe_execute'([File|Files], _) :- !,
	'$cafe_execute'(consult([File|Files]),_).
'$cafe_execute'(help, _) :- !,
	tab(8), write('compile(File).'), nl,
	tab(8), write('consult(File).'), nl,
	tab(8), write('listing.'), nl,
	tab(8), write('trace.'), nl,
	tab(8), write('notrace.'), nl,
	tab(8), write('halt. or ^D'), nl,
	tab(8), write('end_of_file.'), nl.
'$cafe_execute'(Goal, Vs) :-
	'$cafe_exec'(Goal, Vs),
	!,
	write(yes), 
	nl.
'$cafe_execute'(_Goal, _Vs) :-
	nl,
	write(no), 
	nl.

'$cafe_exec'(G, []) :- !, 
	('$trace_flag'(on) -> '$trace_goal'(G) ; '$solve'(G)),
	 nl.
'$cafe_exec'(G, Vs) :-	
	('$trace_flag'(on) -> '$trace_goal'(G) ; '$solve'(G)),
	nl,
	'$cafe_write_result'(Vs), 
	write(' ? '),
	flush_output,
	readline(Str),
        Str \== ";",
	nl.

'$solve'(true) :- !.
'$solve'(!) :- '$ancestor'('$reduce'(_G), L), !, '$cut'(L).
'$solve'(!) :- !.
%'$solve'(!) :- '$ancestor'('$reduce'(G), L), !, '$cut'(L),
%	nl, write('** '), write(cut(G,L)), write(' **'),nl.
'$solve'(not(G)) :- !, \+ '$solve'(G).
'$solve'(\+(G)) :- !, \+ '$solve'(G).
'$solve'(once(G)) :- !, once('$solve'(G)).
'$solve'((IF -> THEN)) :- !,
	'$if_then_else'('$solve'(IF), '$solve'(THEN), fail).
'$solve'((IF -> THEN; ELSE)) :- !, 
	'$if_then_else'('$solve'(IF), '$solve'(THEN), '$solve'(ELSE)).
'$solve'(findall(X,G,Xs)) :- !, 
	findall(X,'$solve'(G),Xs).
'$solve'((G1,G2)) :- !, 
	'$solve'(G1), 
	'$solve'(G2).
'$solve'((G1;G2)) :- !, 
	('$solve'(G1) ; '$solve'(G2)).
'$solve'(call(G)) :- !, '$solve'(G).
'$solve'(G) :- '$call'(G).
'$solve'(G) :- '$reduce'(G).

'$reduce'(G) :- 
	'$get_level'(L), 
	%clause('$consulted_clause'(G, G1), _),
	clause(G, G1),
	('$ancestor'('$reduce'(G), L) => '$solve'(G1)).

'$cafe_write_result'(Vs) :-
	'$remove_redundant_vars'(Vs, Vs1),
	'$cafe_write_result0'(Vs1).

'$remove_redundant_vars'([], []) :- !.
'$remove_redundant_vars'([N=V|Xs], Vs)  :- var(V),!,
	N = V,
	'$remove_redundant_vars'(Xs, Vs).
'$remove_redundant_vars'([X|Xs], [X|Vs]) :-
	'$remove_redundant_vars'(Xs, Vs).

'$cafe_write_result0'([]) :- !, write(true).
'$cafe_write_result0'([N=V]) :- !,
	write(N), write(' = '), write(V).
'$cafe_write_result0'([N=V|Vs]) :-
	write(N), write(' = '), write(V), write(','), nl,
	'$cafe_write_result0'(Vs).

%-----------------------------------------------------------
% consult/1: consulting Prolog programs
%   This is under construction and incomplete,
%   since I don't consider cut(!) any more.
%-----------------------------------------------------------
consult(Files) :- var(Files), !, fail.
consult([]) :- !.
consult(File) :- atom(File), !,
	'$consult'(File).
consult([File|Files]) :- atom(File), !,
	'$consult'(File),
	consult(Files).

'$consult'(FileName) :-
	statistics(runtime, _),
	'$list_to_string'([FileName, '.pl'], In),
	write('{consulting '), write(In), write('...}'), nl,
	'$read_in_programs'(In),
	statistics(runtime, [_,T]),
	write('{'), write(In), write(' consulted, '),
	write(T), write(' msec}'), nl.

%'$consult_init' :-
%	retractall('$consulted_predicates'(_,_)),
%	retractall('$consulted_clause'(_,_)).

'$read_in_programs'(FileName) :- 
	open(FileName, read, In),
	repeat,
	  read(In, X),
	  '$consult_clause'(X),
	X == end_of_file,
	!,
	close(In).

'$consult_clause'(end_of_file) :- !.
'$consult_clause'((:- dynamic _G)) :- !.
%'$consult_clause'((:- dynamic G)) :- !, '$consult_dynamic'(G).
'$consult_clause'((:- mode _G)) :- !.
'$consult_clause'((:- multifile _G)) :- !.
'$consult_clause'((:- public _G)) :- !.
'$consult_clause'((:- block _G)) :- !.
'$consult_clause'((:- meta_predicate _G)) :- !.
'$consult_clause'((:- G)) :- !, '$solve'(G).
'$consult_clause'(Clause0) :-
	'$consult_preprocess'(Clause0, Clause),
	'$consult_cls'(Clause).

'$consult_cls'(Clause) :- assert(Clause).

/*
'$consult_cls'((H :- B)) :- !, 
	'$assert_consulted_predicate'(H),
	assert('$consulted_clause'(H, B)).
'$consult_cls'(H) :- 
	'$assert_consulted_predicate'(H),
	assert('$consulted_clause'(H, true)).

'$assert_consulted_predicate'(Head) :-
	functor(Head, Functor, Arity),
	clause('$consulted_predicates'(Functor, Arity), _),
	!.
'$assert_consulted_predicate'(Head) :-
	functor(Head, Functor, Arity),
	assert('$consulted_predicates'(Functor, Arity)).
*/

'$consult_preprocess'(Clause0, Clause) :-
	expand_term(Clause0, Clause).

%-----------------------------------------------------------
% listing/0: listing current programs
%-----------------------------------------------------------
listing :-
	nl,
	%clause('$consulted_predicates'(F, A), _),
	clause('$current_predicates'(F, A), _),
	name(F, [C|_Cs]),
	C =\= "$",
	functor(Head, F, A),
	%findall((Head :- Body), clause('$consulted_clause'(Head, Body), _), Clauses),
	findall((Head :- Body), clause(Head, Body), Clauses),
	'$write_clauses'(Clauses), nl,
	fail.
listing.

'$write_clauses'([]) :- !.
'$write_clauses'([C|Cs]) :- 
	'$write_clause'(C),
	'$write_clauses'(Cs).

'$write_clause'(C) :-
    numbervars(C, 0, _), 
    '$write_clause0'(C).

'$write_clause0'((H :- B)) :- B == true, !,
    writeq(H),write('.'),nl.
'$write_clause0'((H :- B)) :-
    writeq(H), write(' :- '), nl,
    '$write_body'(B).

'$write_body'(B) :-
	'$list_of_body'(B, Bs, []),
	'$write_body0'(Bs).

'$write_body0'([]) :- !.
'$write_body0'([B]) :- !, 
	tab(8), writeq(B), write('.'), nl.
'$write_body0'([B|Bs]) :-
	tab(8), writeq(B), write(','), nl,
	'$write_body0'(Bs).

'$list_of_body'((B1,B2)) --> !, 
	[B1], '$list_of_body'(B2).
'$list_of_body'(B) --> [B].

%-----------------------------------------------------------
% trace
%-----------------------------------------------------------
'$trace_init' :- 
	retractall('$trace_flag'(_)),
	retractall('$current_leash'(_)),
	retractall('$current_spy'(_,_)),
	assert('$trace_flag'(off)),
	'$full_leash'(Ms),
	'$assert_leashes'(Ms).

debug :- trace.
nodebug :- notrace.

trace :- 
	retract('$trace_flag'(_)),
	assert('$trace_flag'(on)),
	('$current_spy'(_,_) -> 
	     write('{Small debugger is switch on -- show spypoints}')
	;
	     write('{Small debugger is switch on -- show everything}')
	),
	nl,
	!.

notrace :-
	retract('$trace_flag'(_)),
	assert('$trace_flag'(off)),
	write('{Small debugger is switch off}'), 
	nl,
	!.

% spy/1
(spy []) :- !.
(spy X) :- '$spy'(X), trace.

'$spy'(X) :- 
	nonvar(X), 
	!,
	('$create_spy_keys'(X, Keys) -> 
	      '$assert_spies'(Keys)
	;
	      '$illarg'(type(spec), (spy X), 1)
	).
'$spy'(X) :- '$illarg'(var, (spy X), 1).

'$create_spy_keys'(X, Y) :- '$spy_keys'(X, Y).
	
'$spy_keys'([],[]) :- !.
'$spy_keys'(X, [Y]) :- '$spy_key'(X, Y), !.
'$spy_keys'([X|Xs],[Y|Ys]) :- 
	'$spy_key'(X, Y),
	!,
	'$spy_keys'(Xs, Ys).

'$spy_key'(X, X) :- X = F/A, atom(F),integer(A).

'$assert_spies'([]) :- !.
'$assert_spies'([K|Ks]) :-
	'$assert_spy'(K),
	'$assert_spies'(Ks).

'$assert_spy'(K) :- 
	K = F/A,
	clause('$current_spy'(F, A), _), 
	write('{Spypoint '), write(K), write(' is already defined}'), nl,
	!.
'$assert_spy'(K) :- 
	K = F/A,
	assert('$current_spy'(F, A)),
	write('{Spypoint '), write(K), write(' is added}'), nl.

% nospy/1
(nospy []) :- !.
(nospy X) :- '$nospy'(X).

'$nospy'(X) :- 
	nonvar(X), 
	!,
	('$create_spy_keys'(X, Keys) -> 
	      '$remove_spies'(Keys)
	;
	      '$illarg'(type(spec), (nospy X), 1)
	).
'$nospy'(X) :- '$illarg'(var, (nospy X), 1).

'$remove_spies'([]) :- !.
'$remove_spies'([K|Ks]) :-
	'$remove_spy'(K),
	'$remove_spies'(Ks).

'$remove_spy'(K) :- 
	K = F/A,
	retract('$current_spy'(F, A)),
	!,
	write('{Spypoint '), write(K), write(' is removed}'), nl.
'$remove_spy'(_).

% nospyall/1
nospyall :-  retractall('$current_spy'(_,_)).

% leash/1
leash(Mode) :- 
	nonvar(Mode), 
	'$check_leash'(Mode),
	!,
	'$leash'(Mode).
leash(Mode) :- 
	findall(M, '$leash_mode'(M), Domain),
	'$illarg'(domain(atom,Domain), leash(Mode), 1).

'$leash'([]) :- !, 
	retractall('$current_leash'(_)),
	write('{No leashing}'), nl.
'$leash'(none) :- !, '$leash'([]).
'$leash'(full) :- !,
	retractall('$current_leash'(_)),
	'$full_leash'(Ms),
	'$assert_leashes'(Ms),
	write('{Leashing stopping on '), write(Ms), write('}'),nl.
'$leash'(Ms) :-
	retractall('$current_leash'(_)),
	'$assert_leashes'(Ms),
	write('{Leashing stopping on '), write(Ms), write('}'),nl.

'$check_leash'([]) :- !.
'$check_leash'(M) :- '$leash_mode'(M),!.
'$check_leash'([M|Ms]) :- '$leash_mode'(M), '$check_leash'(Ms).

'$assert_leashes'([]) :- !.
'$assert_leashes'([X|Xs]) :-
	assert('$current_leash'(X)),
	'$assert_leashes'(Xs).

'$full_leash'([call,exit,redo,fail]).

'$leash_mode'(call).
'$leash_mode'(exit).
'$leash_mode'(redo).
'$leash_mode'(fail).
'$leash_mode'(full).
'$leash_mode'(none).

%-----------------------------------------------------------
% tracer
%-----------------------------------------------------------
'$trace_goal'(G) :- var(G), !, '$illarg'(var, '$trace_goal'(G), 1).
'$trace_goal'(trace) :- !, trace.
'$trace_goal'(notrace) :- !, notrace.
'$trace_goal'(debug) :- !, debug.
'$trace_goal'(nodebug) :- !, nodebug.
'$trace_goal'(leash(X)) :- !, leash(X).
'$trace_goal'(spy(X)) :- !, spy(X).
'$trace_goal'(nospy(X)) :- !, nospy(X).
'$trace_goal'(nospyall) :- !, nospyall.
'$trace_goal'(G) :- '$trace'(G, 0).

'$trace'(G, Depth) :- var(G), !, '$illarg'(var, '$trace'(G, Depth), 1).
'$trace'(true, _) :- !.
'$trace'(!, Depth) :-
	'$ancestor'('$trace_reduce'(G), L), 
	!, 
	'$tracing_messages'(Depth, !(G, L), call),
	'$cut'(L).
'$trace'(!, _) :- !.
'$trace'(not(G), Depth) :- !, 
	'$trace'(\+(G), Depth).
'$trace'(\+(G), Depth) :- 
	'$trace'(G, Depth), !, fail.
'$trace'(\+(_G), _Depth) :- !.
'$trace'((IF -> THEN), Depth) :- !, 
	'$trace'((IF -> THEN; fail), Depth).
'$trace'((IF -> THEN; _ELSE), Depth) :-
	'$trace'(IF, Depth), !, '$trace'(THEN, Depth).
'$trace'((_IF -> _THEN; ELSE), Depth) :- !,
	'$trace'(ELSE, Depth).
'$trace'(findall(X,G,Xs), Depth) :- !,
	findall(X,'$trace'(G, Depth),Xs).
'$trace'((A,B), Depth) :- !,
	'$trace'(A, Depth), 
	'$trace'(B, Depth).
'$trace'((A;B), Depth) :- !,
	('$trace'(A, Depth) ; '$trace'(B, Depth)).
'$trace'(call(A), Depth) :- !, 
	'$trace'(A, Depth).
'$trace'(A, Depth) :-
	'$tracing_messages'(Depth, A, call),
	%'$write_depth'(Depth), write('Call: '), write(A), write(' ? '), 
	%flush_output,
	%readline(_),
	('$call'(A); '$trace_reduce'(A, Depth)),
	'$tracing_messages'(Depth, A, exit),
	%'$write_depth'(Depth), write('Exit: '), write(A), write(' ? '), 	
	%flush_output,
	%readline(_),
	'$trace_redo'(A, Depth).
'$trace'(A, Depth) :-
	'$tracing_messages'(Depth, A, fail),
	%'$write_depth'(Depth), write('Fail: '), write(A), write(' ? '), 	
	%flush_output,
	%readline(_),
	fail.

'$trace_reduce'(A, Depth) :- 
	'$get_level'(L), 
	%clause('$consulted_clause'(A, B), _),
	clause(A, B),
	Depth1 is Depth + 1,
	('$ancestor'('$trace_reduce'(A), L) => '$trace'(B, Depth1)).

'$trace_redo'(_, _).
'$trace_redo'(A, Depth) :- 
	'$tracing_messages'(Depth, A, redo),
	%'$write_depth'(Depth), write('Redo: '), write(A), write(' ? '), 
	%flush_output,
	%readline(_),
	fail.

'$write_depth'(0) :- !, '$fast_write'(' + ').
'$write_depth'(N) :- 
	N > 0,
	'$fast_write'(' |'),
	M is N-1,
	'$write_depth'(M).

'$tracing_messages'(Depth, Goal, Mode) :- 
	'$current_spy'(_, _),
	!,
	'$spy_messages'(Depth, Goal, Mode).

'$tracing_messages'(Depth, Goal, Mode) :- 
	'$current_leash'(Mode),
	!,
	'$write_depth'(Depth), 
	'$write_leash_mode'(Mode),
	write(Goal), 
	write(' ? '), flush_output, readline(_).
%'$tracing_messages'(_Depth, _Goal, _Mode).
'$tracing_messages'(Depth, Goal, Mode) :- 
	'$write_depth'(Depth), 
	'$write_leash_mode'(Mode),
	write(Goal), nl.

'$spy_messages'(Depth, Goal, Mode) :-
	functor(Goal, Functor, Arity),
	'$current_spy'(Functor, Arity),
	!,
	write(' + '), write(Depth), write(' '),
	'$write_leash_mode'(Mode),
	write(Goal),
	write(' ? '), flush_output, readline(_).
'$spy_messages'(_Depth, _Goal, _Mode).

'$write_leash_mode'(call) :- !, write('Call: ').
'$write_leash_mode'(exit) :- !, write('Exit: ').
'$write_leash_mode'(redo) :- !, write('Redo: ').
'$write_leash_mode'(fail) :- !, write('Fail: ').

%-----------------------------------------------------------
% Utilities
%-----------------------------------------------------------

%'$range'(I, N, []) :- I > N, !.
%'$range'(I, N, [I|L]) :- I =< N, I1 is I+1, '$range'(I1, N, L).





