/*
 *  @(#)CsObjectTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */
// import java.lang.*;

namespace JJC.Psharp.Lang {

	using System;
	using System.Collections;

	public sealed class CsObjectTerm : Term {
		public static readonly SymbolTerm Nil = SymbolTerm.makeSymbol("[]");
		private object obj;
		private Type type; // CHANGE TO A TYPE, WAS Class

		public CsObjectTerm(object o) { 
			obj = o; 
			type = obj.GetType( ); // obj.getClass(); TOOK OUT
		}

		public void undo() { obj = null; }

		public void set(object o) { 
			obj = o; 
			type = obj.GetType( ); // obj.getClass(); TOOK OUT
		}

		public object Object() { return obj; }

		public override bool unify(Term t, Trail trail) {
			t = t.dereference();
			if ( t.isVariable() )
				return ((VariableTerm)t).unify(this, trail);
			if ( !t.isCsObject() )
				return false;
			return obj.Equals(((CsObjectTerm)t).Object());
		}

		public override bool Equals(Term t) {
			//	t = t.dereference();
			if( !t.isCsObject() )
				return false;
			return obj.Equals(((CsObjectTerm)t).Object());
		}

		//    public Term copyTerm(Hashtable table) {
		//	return new CsObjectTerm(obj);
		//    }

		public override Term copyTerm(Prolog engine) {
			return new CsObjectTerm(obj);
		}

		public override object toCsObject() {
			return obj;
		}

		public Term toPrologTerm() {
			return csTerm(obj);
		}

		private Term csTerm(object o) {
			if( o is Int32 )
				return new IntegerTerm( (Int32)o );
			if( o is string )
				return SymbolTerm.makeSymbol( (string)o );
			if( o is Boolean )
				return SymbolTerm.makeSymbol( o.ToString( ) );
			if( o is Array ) {
				object[] oo = (Object[])o;
				Term t = Nil;
				foreach( object ooo in oo )
					t = new ListTerm( (Term)ooo, t );
				return t;
			}
			return new CsObjectTerm( o );
		
			/* DO LATER 
			 * 
				Class c = o.getClass();
			if ( c.equals(Integer.class) )
				return new IntegerTerm(((Integer)o).intValue());
			if ( c.equals(String.class) )
				return SymbolTerm.makeSymbol((String)o);
			if ( c.equals(Boolean.class) )
				return SymbolTerm.makeSymbol(o.ToString());
			if ( c.isArray() ){
				Object[] oo = (Object[])o;
				Term t = Nil;
				for (int i=oo.Length; i>0; i--){
				t = new ListTerm(javaTerm(oo[i-1]), t);
				}
				return t;
			}
			*/
			//	return new CsObjectTerm(o);
		}

		/*
		public Class getJavaType(){
		return clazz;
		}*/

		public Type getType() { // was Class
			return type;
		}

		public override string ToString() {
			if (obj == null)
				return "CsObjectTerm(null)";
			return type.FullName + "(" + obj.GetHashCode( ) + ")";
		}
	}

}