/*
 * @(#)PrologInterface.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Threading;

	public sealed class PrologInterface {
		private bool result;
		Predicate code = null;
		Prolog engine = new Prolog();

		public void setPredicate(Predicate p) {
			code = p;
		}

		public bool call() {
			lock( this ) {
				this.start();	 /* execute run() in new thread */
				try {
					Monitor.Wait(this);	     /* wait caller's thread */
				} catch (ThreadInterruptedException) {}
				return result;
			}
		}

		public bool redo() {
			lock( this ) {
				Monitor.Pulse(this);	    /* resume child thread */
				try {
					Monitor.Wait(this);     /* wait caller's thread */
				} catch (ThreadInterruptedException) {}
				return result;
			}
		}

		public void success() {
			lock( this ) {					   
				result = true;
				Monitor.Pulse(this);   	/* resume caller's thread */
				try {
					Monitor.Wait(this); 	/* wait child thread */
				} catch (ThreadInterruptedException) {}
			}
		}

		public void fail() {
			lock( this ) {
				result = false;
				Monitor.Pulse(this);	    /* resume caller's thread */
			}
		}

		public void start() { // EXPERIMENT.
			new Thread( new ThreadStart( run ) ).Start( );
		}

		public void run() {
			try {
				while (true) {
					if (code == null) break;
					// System.Console.WriteLine( code.ToString( ) );
					code.setEngine(engine);
					code = code.exec();
				}
			}
			catch (AbortException e){
				System.Console.WriteLine(e.Message);
			}
			fail();
		}
	}
}