/*
 * @(#)StructureTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

// import java.util.*;

namespace JJC.Psharp.Lang {

	public sealed class StructureTerm : Term {
		private SymbolTerm ifunctor;
		public SymbolTerm functor {
			get {
				return ifunctor;
			}
			set {
				ifunctor = value;
			}
		}

		private Term[] iargs;
		public Term[] args {
			get {
				return iargs;
			}
			set {
				iargs = value;
			}
		}

		private int iarity;
		public int arity {
			get {
				return iarity;
			}
			set {
				iarity = value;
			}
		}

		public StructureTerm(SymbolTerm _functor, Term[] _args){
			functor = _functor;
			args = _args;
			arity = functor.arity;
		}

		public string name(){
			return functor.name;
		}

		public override Term copyTerm(Prolog engine) {
			Term[] a = new Term[arity];
			for (int i = 0; i < arity; i++)
				a[i] = args[i].copyTerm(engine);
			return new StructureTerm(functor, a);
		}

		/*
		public Term copyTerm(Hashtable table) {
		Term[] a = new Term[arity];
		for (int i = 0; i < arity; i++)
			a[i] = args[i].copyTerm(table);
		return new StructureTerm(functor, a);
		}
		*/

		public override bool unify(Term t, Trail trail) {
			t = t.dereference();
			if ( t.isVariable() )
				return ((VariableTerm)t).unify(this, trail);
			if ( !t.isStructure() )
				return false;
			StructureTerm s = (StructureTerm)t;
			if ( functor != s.functor )
				return false;
			for (int i=0; i<arity; i++)
				if ( !args[i].unify(s.args[i], trail) )
					return false;
			return true;
		}

		public override bool Equals(Term t) {
			Term a1, a2;
			//	t = t.dereference();
			if ( !t.isStructure() )
				return false;
			StructureTerm s = (StructureTerm)t;
			if ( functor != s.functor )
				return false;
			for ( int i = 0; i < arity; i++ ){
				a1 = args[i].dereference();
				a2 = s.args[i].dereference();
				if ( ! a1.Equals(a2) )
					return false;
			}
			return true;
		}

		public override string ToString() {
			string s;
			string delim = "";
			s = functor.ToString() + "(";
			for (int i = 0; i < arity; i++) {
				s += delim + args[i].ToString();
				delim = ",";
			}
			s += ")";
			return s;
		}
	}

}