/*
 * @(#)Trail Stack
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Collections;

	using JJC.Psharp.Lang.Resource;

	public sealed class Trail {
		private int maxContents = 20000;
		private Stack buffer;

		private int itop;
		public int top {
			get {
				return itop;
			}

			set {
				itop = value;
			}
		}
		internal Prolog engine;
	
		public Trail() {
			buffer = new Stack( maxContents ); // Term[maxContents];
			top = -1;
		}
		public Trail(int n){
			maxContents = n;
			buffer = new Stack( maxContents ); // Term[maxContents];
			top = -1;
		}

		public void setEngine(Prolog _engine){
			engine = _engine;
		}

		public void push(Term t){

			++top;
			buffer.Push( t );
			/*
			   try { 
					buffer[++top] = t;
				} catch (IndexOutOfRangeException) {
					Console.WriteLine("{expanding trail...}");
					int len = buffer.Length;
					Term[] new_buffer = new Term[len+20000];
					for(int i=0; i<len; i++){
						new_buffer[i] = buffer[i];
					}
					buffer = new_buffer;
					buffer.Peek() = t;
				}
			*/	
	
		}

		public Term pop(){
			Term t = (Term)buffer.Pop( );
			top--; 
			// buffer[top--] = null;
			return t;
		}

		public bool empty(){
			return top == -1;
		}

		public int max(){
			return maxContents;
		}

		public void unwind(int i){
			Term t, s;
			String key;
			int n;
			int k,m;
	
			while ( top > i ){
				t = pop();
				if ( t.isVariable() ){
					((VariableTerm)t).undo();
					continue;
				}
				if ( t.isInteger() ){
					n = ((IntegerTerm)t).value();
					k = ((ResIndexTerm)pop()).getIndex();
					engine.res.setLevel(k,n);
					continue;
				}
				if ( t.isSymbol() ) {
					((SymbolTerm)t).undoRes();
					s = pop();
					if (! s.isSymbol() ) Console.WriteLine("are????");
					if ( s == t )
						((SymbolTerm)s).undoRes2();
					else{
						key = ((SymbolTerm)s).name;
						t = (Term)engine.res.hash[key];
						if (! t.isList() ) Console.WriteLine("are????");
						engine.res.hash[ key ] = ((ListTerm)t).cdr.dereference();
					}
					continue;
				}
				if ( t.isResIndex() ){
					m = ((ResIndexTerm)t).getIndex();
					k = ((ResIndexTerm)pop()).getIndex();
					for (; k<m; k++){
						engine.res.setOutOfScope(k, false);
					}
					continue;
				}
				Console.WriteLine("error in unwind trail");
				//	    engine.abort("");
			}
		}

		public void show(){
			if ( empty() ) {
				Console.WriteLine("Trail is empty!");
				return;
			}
			Console.WriteLine( this.ToString( ) );
			/*
			for ( int i=0; i<=top; i++ ){
				Console.WriteLine("trail[" + i.ToString() + "]: ");
				Console.WriteLine(buffer[i]);
			}
			*/
		}
	}

}