/*
 * @(#)VariableTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

// import java.util.*;

namespace JJC.Psharp.Lang {

	using System;

	public sealed class VariableTerm : Term {
		private static long counter = Int64.MinValue;
		//private static long count = 0;
		//private Term val;
		Term val;

		private long itimeStamp;
		public long timeStamp {
			get {
				return itimeStamp;
			}
			set {
				itimeStamp = value;
			}
		}
		private long ID;

		public VariableTerm(long t) {
			//	val = null;
			val = this;
			timeStamp = t;
			ID = counter++;
		}

		public VariableTerm() {
			//	val = null;
			val = this;
			timeStamp = Int64.MinValue;
			ID = counter++;
		}

		public override Term dereference() {
			if (val == this)
				return this;
			return val.dereference();
		}

		//    public void undo() { val = null; }
		public void undo() { val = this; }

		/*
		public Term copyTerm(Hashtable table) {
		VariableTerm copy;
		if (val != null)
			return val.copyTerm(table);
		copy = (VariableTerm)table.get(this);
		if ( copy == null ){
			copy = new VariableTerm();
			table.put(this, copy);
		}
		return copy;
		}
		*/

		public override Term copyTerm(Prolog engine) {
			System.Collections.Hashtable table = engine.varTable;
			VariableTerm copy;
			if (val != this)
				return val.copyTerm(engine);
			copy = (VariableTerm)table[this];
			if ( copy == null ){
				copy = engine.makeVariable();
				table[ this ] = copy;
			}
			return copy;
		}

		/*
		public bool unify(Term t, Trail trail) {
		if (val != null)
			return val.unify(t, trail);
		t = t.dereference();
		if (this != t) {
			val = t;
			trail.push(this);
		}
		return true;
		}*/

		public override bool unify(Term t, Trail trail) {
			VariableTerm v1;
			if (val != this)
				return val.unify(t, trail);
			t = t.dereference();
			if ( t.isVariable() ){
				VariableTerm v2 = (VariableTerm)t;
				if (this.timeStamp < v2.timeStamp ){
					v2.val = this;
					v1 = v2;
				} else {
					this.val = t;
					v1 = this;
				}
			} else {
				this.val = t;
				v1 = this;
			}
			if ( v1.timeStamp < trail.engine.stack.getTimeStamp() ){
				trail.push(v1);
			}
			return true;
		}

		public override bool Equals(Term t) {
			if( val != this )
				return val.Equals(t);
			//	t = t.dereference();
			if ( !t.isVariable() )
				return false;
			return this == t;
			//	return this.timeStamp == ((VariableTerm)t).timeStamp();
		}

		/*
		public Class getJavaType() { 
		if (val != null)
			return val.getJavaType();
		return java.lang.Object.class;
		}*/
		public override object toCsObject() {
			if (val != this)
				return val.toCsObject();
			return this;
			//	return new CsObjectTerm(this);
		}

		public String name() {
			//return "_" + timeStamp();
			return "_" + this.GetHashCode( ).ToString( );
		}

		public long getID() {
			return ID;
		}

		public override String ToString() {
			if (val != this)
				return val.ToString();
			return name();
		}
	}

}